/*
   Copyright (C) 2005 Michael Nelson Moran

   This file is part of OSCL.

   OSCL is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   OSCL is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.

   You should have received a copy of the GNU General Public License
   along with OSCL; see the file COPYING.  If not, write to the Free
   Software Foundation, 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   To negotiate other licensing arrangments/agreements contact
   the copyright holder:

   Michael N. Moran
   5009 Old Field Ct.         mike@mnmoran.org
   Kennesaw, GA, USA 30144    http://mnmoran.org
*/

#include "ring.h"
#include "oscl/endian/type.h"
#include "oscl/cpu/sync.h"

using namespace Oscl::Buffer::Ring;

TxDesc::TxDesc(	SharedApi&	shared,
				Unsigned32	waterMark
				) throw():
		_allocationOffset(0),
		_waterMark(waterMark),
		_nCommits(0),
		_shared(shared)
		{
	while(!_shared.isReady());

	Unsigned32	maxFifoDepth	= _shared.maxFifoDepth();

	if(waterMark > maxFifoDepth){
		_waterMark	= maxFifoDepth;
		}
	}

bool	TxDesc::alloc(Unsigned32& index) throw(){

	Unsigned32
	nElements	= _shared.nElements();

	Unsigned32
	writeIndex	= _shared.writeIndex() + _allocationOffset;

	writeIndex	%=	nElements;


	Unsigned32
	nextIndex	= (writeIndex + 1) % nElements;

	if( nextIndex == _shared.readIndex() ) {
		return false;
		}
	else {
		index	= writeIndex;
		++_allocationOffset;
		return true;
		}
	}

bool	TxDesc::isFull() const throw(){
	Unsigned32
	nElements	= _shared.nElements();

	Unsigned32
	index	= _shared.writeIndex() + _allocationOffset;

	index	%=	nElements;

	Unsigned32
	nextIndex	=	(index + 1) % nElements;

	return (nextIndex == _shared.readIndex());
	}

void	TxDesc::commit() throw(){
	Unsigned32	nCommits;

	if(_waterMark) {
		++_nCommits;
		if(_waterMark < _nCommits) {
			return;
			}
		nCommits	= _nCommits;
		_nCommits	= 0;
		_waterMark	= 0;
		}
	else {
		nCommits	= 1;
		}

	Unsigned32
	index	= _shared.writeIndex();

	Unsigned32
	nextIndex	= (index + nCommits) % _shared.nElements();

	OsclCpuInOrderMemoryAccessBarrier();

	_shared.writeIndex(nextIndex);

	_allocationOffset	-= nCommits;
	}

RxDesc::RxDesc(	SharedApi&	shared
				) throw():
		_allocationOffset(0),
		_shared(shared)
		{
	}

bool	RxDesc::alloc(Unsigned32& index) throw(){
	Unsigned32
	readIndex	= _shared.readIndex() + _allocationOffset;

	Unsigned32
	nElements	= _shared.nElements();

	readIndex		%= nElements;

	Unsigned32
	writeIndex	= _shared.writeIndex();

	if( readIndex == writeIndex ) {
		return false;
		}
	else {
		++_allocationOffset;
		index	= readIndex;
		return true;
		}
	}

void	RxDesc::commit() throw(){
	Unsigned32
	nElements	= _shared.nElements();

	Unsigned32
	index		=	_shared.readIndex() + 1;
	index		%=	nElements;

	OsclCpuInOrderMemoryAccessBarrier();

	_shared.readIndex(index);

	--_allocationOffset;

	}

